/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  $Header: /home/ogg/repository/ogg/apps/appio.c,v 1.3 1998/03/04 07:05:45 ogg Exp $
  $Author: ogg $
  $Date: 1998/03/04 07:05:45 $
  modifications by: Monty
  function: Application top level

 ********************************************************************/

#include "oggapp.h"
#include <sys/stat.h>

#ifndef lint
static char rcs_id[] = "$Id: appio.c,v 1.3 1998/03/04 07:05:45 ogg Exp $";
#endif

extern int OGG_verbose;
extern int OGG_X;

void reporterror(char *s){
  fprintf(stderr,"\nERROR: %s\n",s);
  exit(-1);
}

void flatreport(char *s){
  if(OGG_verbose && !OGG_X){
    fprintf(stderr,"%s",s);
    fflush(stderr);
  }
}

void report(char *s){
  flatreport(s);
  flatreport("\n");
}

char *static_buffer=NULL;      /* When we read headers from a       */
long static_buffer_size=0;     /* non-seekable stream, more than one */
long static_buffer_alloced=0;  /* function will need to check the same
			       data; we can't put it back, so we need
			       to buffer it */

int query_seekable(FILE *f){
  int fd;
  struct stat buf;

  fd=fileno(f);
  fstat(fd,&buf);
  if(buf.st_mode&S_IFREG)return(-1);
  return(0);
}

static char **buffer_cache;
static long *bytes_cache;
static long *alloced_cache;
static ino_t *inode_cache;
static long cachesize;

FILE *cached_open(char *filename,char *mode){
  struct stat buf;
  int fd,i,flag=-1;
  FILE *ret=fopen(filename,mode);
  if(ret==NULL)return(NULL);

  fd=fileno(ret);
  fstat(fd,&buf);
  for(i=0;i<cachesize;i++)
    if(inode_cache[i]==buf.st_ino){
      flag=i;
      break;
    }

  if(flag!=-1){
    /* It's a-cach'ed babeee... */
    /* That's it... decaf the rest of the night */
    static_buffer=buffer_cache[flag];
    static_buffer_size=bytes_cache[flag];
    static_buffer_alloced=alloced_cache[flag];
    
    for(i=flag;i<cachesize-1;i++){
      buffer_cache[i]=buffer_cache[i+1];
      alloced_cache[i]=alloced_cache[i+1];
      bytes_cache[i]=bytes_cache[i+1];
      inode_cache[i]=inode_cache[i+1];
    }
    cachesize--;
  }
  return(ret);
}

/* Cache the bytes in the static buffer if we're not a regular file */

void cached_close(FILE *f){
  if(!query_seekable(f)){
    struct stat buf;
    int fd,i,flag=-1;

    fd=fileno(f);
    fstat(fd,&buf);
    for(i=0;i<cachesize;i++)
      if(inode_cache[i]==buf.st_ino){
	flag=i;
	break;
      }
    
    if(flag!=-1)
      /* Oh; we can replace an entry */
      free(buffer_cache[flag]);
    else{
      flag=cachesize++;
      if((buffer_cache=realloc(buffer_cache,sizeof(char *)*cachesize))==NULL)
	reporterror("Unable to alloc memory.");
      if((alloced_cache=realloc(alloced_cache,sizeof(long *)*cachesize))==NULL)
	reporterror("Unable to alloc memory.");
      if((bytes_cache=realloc(bytes_cache,sizeof(long *)*cachesize))==NULL)
	reporterror("Unable to alloc memory.");
      if((inode_cache=realloc(inode_cache,sizeof(ino_t *)*cachesize))==NULL)
	reporterror("Unable to alloc memory.");
    }
     
    buffer_cache[flag]=static_buffer;
    alloced_cache[flag]=static_buffer_alloced;
    bytes_cache[flag]=static_buffer_size;
    inode_cache[flag]=buf.st_ino;
    static_buffer=NULL;
    static_buffer_alloced=0;
    static_buffer_size=0;
  }else
    if(static_buffer)free(static_buffer);

  static_buffer=NULL;
  static_buffer_alloced=0;
  static_buffer_size=0;
  fclose(f);
}

void cached_close_flush(FILE *f){
  if(static_buffer)free(static_buffer);

  static_buffer=NULL;
  static_buffer_alloced=0;
  static_buffer_size=0;
  fclose(f);
}

/* static buffer hack maint */
/* clean pipe behavior */

long blocking_read(char *buffer, long size, long num, FILE *inf){
  long words=0;

  while (words<num){
    if(feof(inf))break;
    errno=0;
    BOUNDS_CHECKING_OFF;
    words+=fread(buffer+words,size,num-words,inf);
    BOUNDS_CHECKING_ON;
    switch(errno){
    case 0:
    case EINTR:
      break;
    default:
      report("\nERROR while reading input stream:");
      reporterror(strerror(errno));
      break;
    }
    if(OGG_verbose)
      if(words<num && !feof(inf))
	flatreport("\r=[ Waiting for input ]==>");
      
  }
  return(words);
}

long blocking_write(char *buffer, long size, long num, FILE *outf){
  long words=0;

  while (words<num){
    if(feof(outf))break;
    errno=0;
    words+=fwrite(buffer+words,size,num-words,outf);
    switch(errno){
    case 0:
    case EINTR:
      break;
    default:
      report("\nERROR while writing output stream:");
      reporterror(strerror(errno));
      break;
    }
    if(OGG_verbose)
      if(words<num && !feof(outf))
	flatreport("\r==[ Waiting to write ]==>");
      
  }
  return(words);
}

long fd_blocking_write(char *buffer, long num, int outf){
  long words=0,temp;

  while(words<num){
    temp=write(outf,buffer+words,num-words);
    if(temp==-1){
      switch(errno){
      case 0:
      case EINTR:
	temp=0;
	break;
      default:
	report("\nERROR while writing output stream:");
	reporterror(strerror(errno));
      }
    }
    words+=temp;
    if(OGG_verbose)
      if(words<num)
	flatreport("\r==[ Waiting to write ]==>");
    
  }
  return(words);
}

void static_buffer_verify(long bytes){
  if(bytes>static_buffer_alloced){
    if(static_buffer_alloced>0)
      static_buffer=(char *)realloc(static_buffer,bytes);
    else
      static_buffer=(char *)malloc(bytes);
    if(static_buffer==NULL)reporterror("Unable to malloc.");
    static_buffer_alloced=bytes;
  }
}

void static_buffer_read(FILE *f,long bytes){
  if(static_buffer_size==0 && bytes && feof(f))
    reporterror("Unexpected end of data!");
  if(bytes>static_buffer_size){
    int readb;

    /* alloc more space if necessary */
    static_buffer_verify(bytes);

    /* read data into buffer */
    readb=blocking_read(static_buffer+static_buffer_size,1,
			bytes-static_buffer_size,f);
    static_buffer_size+=readb;
  }
}

void static_buffer_shift(long bytes){
  if(bytes>=static_buffer_size)
    static_buffer_size=0;
  else{
    memmove(static_buffer,static_buffer+bytes,static_buffer_size-bytes);
    static_buffer_size-=bytes;
  }
}

