/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  $Id: linux.c,v 1.2 1998/01/20 04:55:17 ogg Exp $
  $Date: 1998/01/20 04:55:17 $
  function: Playback code for Linux and OSS
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty, Wrong Mike

 ********************************************************************/

/* TODO:

   play with small fragments/ better banking to give us the most time
   possible between blocking to process a frame */

/* Can we set a playback mode exactly matching the sample? */
/* If not, set the best we can, or use transforms and effects to force
   fit the playback to the hardware */

#if defined (__linux__) || defined (OPEN_SOUND_SYSTEM)

#include <sys/soundcard.h>
#include <sys/ioctl.h>

int audio=0;

extern int OGG_ess;
extern long OGG_polyphonic;
extern long OGG_mulaw;
extern long O_channels;
extern long O_rate;
extern long O_bits;
extern long O_bigendian;
extern long O_signed;

extern long I_channels;
extern long I_rate;
extern long I_bits;
extern long I_bigendian;
extern long I_signed;

extern int OGG_quality;
extern int OGG_time;

extern long OGG_begin;
extern long OGG_end;
extern long OGG_frames;

extern int OGG_in;  /* 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
extern int OGG_out; /* -1 playback, 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
extern int OGG_action; /* 0 encode, 1 decode, 2 play */

extern int X_channel;
extern int X_resample;
extern int X_volume;

void setplayback(char*);
void closeplayback();
void killplayback();
void playback(void*, long, long);

#ifdef HAVE_LIBPTHREAD
#include <pthread.h>

void playback_consumer();

#define OGG_OUTPUT_RING_SIZE 1572864

pthread_t consumer_id;
pthread_mutex_t ring_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t ring_read_cv=PTHREAD_COND_INITIALIZER;
pthread_cond_t ring_write_cv=PTHREAD_COND_INITIALIZER;

void **handle_ring=NULL;
long *size_ring=NULL;
int ring_read_idx=0, ring_write_idx=0;
int ring_elements=0, elements_stored=0;
int consumer_cmd=0; /* 0=go; 1=done; 2=die */

#endif

void setplayback(char *outfilename){
  char buffer[160];
  int temp,temp2;

  if(outfilename==NULL){
    /* Try to open /dev/dsp, /dev/dsp1, /dev/dsp2, /dev/audio */
    if((audio=open("/dev/dsp",O_WRONLY | O_EXCL))==-1)
    if((audio=open("/dev/dsp1",O_WRONLY | O_EXCL))==-1)
    if((audio=open("/dev/dsp2",O_WRONLY | O_EXCL))==-1)
    audio=open("/dev/audio",O_WRONLY | O_EXCL);
  }else
    audio=open(outfilename,O_WRONLY | O_EXCL);

  if(audio==-1)
    reporterror("Unable to open audio playback device.");

  /* Set format */
  /* Go for exact first */
  switch(O_bits){
  case -8:
    temp=AFMT_MU_LAW;
    break;
  case 8:
    temp=AFMT_U8;
    O_signed=0;
    break;
  default:
    temp=AFMT_S16_LE;
    O_signed=-1;
    O_bigendian=OGG_LITTLEENDIAN;
    O_bits=16;
    break;
  }

  temp2=temp;
  if(ioctl(audio,SNDCTL_DSP_SETFMT,&temp)==-1 || temp!=temp2){
    /* Oops, no good...  Let's try a bunch of formats till we can set one.
       Libogg can do these conversions trivially */

    temp=AFMT_S16_LE;
    O_signed=-1;
    O_bigendian=OGG_LITTLEENDIAN;
    O_bits=16;
    temp2=temp;
    if(ioctl(audio,SNDCTL_DSP_SETFMT,&temp)==-1 || temp2!=temp){
      temp=AFMT_S16_BE;
      O_signed=-1;
      O_bigendian=OGG_BIGENDIAN;
      O_bits=16;
      temp2=temp;
      if(ioctl(audio,SNDCTL_DSP_SETFMT,&temp)==-1 || temp2!=temp){
	temp=AFMT_U8;
	O_signed=0;
	O_bigendian=OGG_LITTLEENDIAN;
	O_bits=8;
	temp2=temp;
	if(ioctl(audio,SNDCTL_DSP_SETFMT,&temp)==-1 || temp2!=temp)
	  /* Oh, boy...  Die miserably... */
	  reporterror("Unable to set any common playback format");
	else
	  report("Best available playback mode is 8 bit");
      }else
	report("Using 16 bit playback");
    }else
      report("Using 16 bit playback");
  }

  /* Well, OK, format is set... */

  /* Set channels */

  switch(O_channels){
  case 1:
    temp=0;
    if(ioctl(audio,SNDCTL_DSP_STEREO,&temp)==-1 || temp!=0)
      reporterror("Unable to set monophonic playback mode.");
    break;
  case 2:
    temp=1;
    if(ioctl(audio,SNDCTL_DSP_STEREO,&temp)==-1 || temp!=1){
      /* Ack... try to set mono */
      temp=0;
      if(ioctl(audio,SNDCTL_DSP_STEREO,&temp)==-1 || temp!=0)
	reporterror("Unable to set stereo or mono playback mode.");
      O_channels=1;
      X_channels=1;
      report("Stereo playback unavailable; using mono.");
    }
    break;
  default:
    temp=1;
    if(ioctl(audio,SNDCTL_DSP_STEREO,&temp)==-1 || temp!=1){
      /* Ack... try to set mono */
      temp=0;
      if(ioctl(audio,SNDCTL_DSP_STEREO,&temp)==-1 || temp!=1)
	reporterror("Unable to set stereo or mono playback mode.");
      O_channels=1;
      X_channels=1;
      report("Using mono playback");
    }else{
      report("Using stereo playback");
      O_channels=2;
      X_channels=2;
    }
    break;
  }
      
  temp=O_rate;
  if(ioctl(audio,SNDCTL_DSP_SPEED,&temp)==-1)
    reporterror("Error setting playback speed");
  
  if(fabs(temp-O_rate)/O_rate<.01)temp=O_rate;

  if(temp!=O_rate){
    sprintf(buffer,"Unable to set playback speed of %ldHz. Resampling to %ldHz.",
	    (long)O_rate,(long)temp);
    report(buffer);
    X_resample=O_rate=temp;
  }

#ifdef HAVE_LIBPTHREAD
  pthread_create(&consumer_id, NULL, (void*) playback_consumer, NULL);
#endif

}

void closeplayback(void){
#ifdef HAVE_LIBPTHREAD
  pthread_mutex_lock(&ring_mutex);
  consumer_cmd=1;
  pthread_cond_signal(&ring_read_cv);
  pthread_mutex_unlock(&ring_mutex);
  pthread_join(consumer_id, NULL);
#else
  close(audio);
#endif
}

void killplayback(void){
#ifdef HAVE_LIBPTHREAD
  pthread_mutex_lock(&ring_mutex);
  consumer_cmd=2;
  pthread_cond_signal(&ring_read_cv);
  pthread_mutex_unlock(&ring_mutex);
  pthread_join(consumer_id, NULL);
#else
  ioctl(audio,SNDCTL_DSP_RESET);
  closeplayback();
#endif
}

void playback(void *buffer, long bytes, long bytesperframe){
  
  if(OGG_ess && O_bits==16){
    /* The ESS AudioDrive has a bigtime zero crossing bug; change
       the data into 15 nonnegative bits.  Playback is better
       (much less distortion) although we lose ~6dB SNR */

    long i,els;
    els=bytes>>1;
    for(i=0;i<els;i++)
      ((size16 *)buffer)[i]=(((size16 *)buffer)[i]>>1)+0x4000;
  }

#ifdef HAVE_LIBPTHREAD
  pthread_mutex_lock(&ring_mutex);

  if(ring_elements==0) {
    ring_elements=OGG_OUTPUT_RING_SIZE/bytes;
    if((handle_ring=malloc(ring_elements*sizeof(void*)))==NULL)
      reporterror("Unable to malloc.");
    if((size_ring=malloc(ring_elements*sizeof(long*)))==NULL)
      reporterror("Unable to malloc.");
    ring_read_idx=ring_write_idx=elements_stored=0;
    /* debug_msg(stderr, "producer: allocated ring buffers: %d elements\n", 
	      ring_elements); */
  }

  if(elements_stored>=ring_elements/2){
    pthread_cond_signal(&ring_read_cv);
    /* debug_msg(stderr, "producer: signalling consumer\n"); */
  }
  while(elements_stored>=ring_elements){
    /* debug_msg(stderr, "producer: waiting on full ring buffer\n"); */
    pthread_cond_wait(&ring_write_cv, &ring_mutex);
  }

  elements_stored++;
  /* debug_msg(stderr, "producer: writing ring index %d\n", ring_write_idx); */
  handle_ring[ring_write_idx]=buffer;
  size_ring[ring_write_idx]=bytes;
  /* debug_msg(stderr, "\tbuffer = %p\tsize = %ld\telements_stored = %d\n", 
	    buffer, bytes, elements_stored); */
  if(++ring_write_idx==ring_elements) ring_write_idx=0;

  pthread_mutex_unlock(&ring_mutex);
#else
  fd_blocking_write(buffer,bytes,audio);
#endif
}

#ifdef HAVE_LIBPTHREAD
void playback_consumer(){
  void *buffer;
  long bytes;

  pthread_mutex_lock(&ring_mutex);
  while(ring_elements==0) {
    /* debug_msg(stderr, "consumer: waiting for initial wakeup\n"); */
    pthread_cond_wait(&ring_read_cv, &ring_mutex);
  }
  /* debug_msg(stderr, "consumer: initual wakeup\n"); */
  pthread_mutex_unlock(&ring_mutex);

  for(;;){
    pthread_mutex_lock(&ring_mutex);

    if(elements_stored==0 && consumer_cmd==0)
      while(elements_stored<ring_elements/2 && consumer_cmd==0){
	/* debug_msg(stderr, "consumer: waiting on empty ring buffer\n"); */
	pthread_cond_wait(&ring_read_cv, &ring_mutex);
      }

    if(consumer_cmd==2){
      pthread_mutex_unlock(&ring_mutex);
      ioctl(audio,SNDCTL_DSP_RESET);
      close(audio);
      break;
    }

    if(elements_stored==0){
      pthread_mutex_unlock(&ring_mutex);
      close(audio);
      break;
    }
    
    /* debug_msg(stderr, "consumer: reading ring index %d\n", ring_read_idx); */
    buffer=handle_ring[ring_read_idx];
    bytes=size_ring[ring_read_idx];
    /* debug_msg(stderr, "\tbuffer = %p\tsize = %ld\telements_stored = %d\n", 
	      buffer, bytes, elements_stored); */
    if(++ring_read_idx==ring_elements) ring_read_idx=0;
    if(elements_stored--==ring_elements){
      /* debug_msg(stderr, "consumer: signalling producer\n"); */
      pthread_cond_signal(&ring_write_cv);
    }
    
    pthread_mutex_unlock(&ring_mutex);

    /* debug_msg(stderr, "consumer: playing block\n"); */
    fd_blocking_write(buffer,bytes,audio);
    free(buffer);
  }
}
#endif  /* HAVE_LIBPTHREAD */

#endif
