/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: sgi.c
  function: Playback code for SGI using AL
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Oct 24 1996

 ********************************************************************/

/* Can we set a playback mode exactly matching the sample? */
/* If not, set the best we can, or use transforms and effects to force
   fit the playback to the hardware */

#ifdef __sgi
#include <dmedia/audio.h>
ALport outaudio;

extern long OGG_polyphonic;
extern long OGG_mulaw;
extern long O_channels;
extern long O_rate;
extern long O_bits;
extern long O_bigendian;
extern long O_signed;

extern long I_channels;
extern long I_rate;
extern long I_bits;
extern long I_bigendian;
extern long I_signed;

extern int OGG_quality;
extern int OGG_time;

extern long OGG_begin;
extern long OGG_end;
extern long OGG_frames;

extern int OGG_in;  /* 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
extern int OGG_out; /* -1 playback, 0 OggSquish, 1 Raw, 2 WAVE, 3 AIFF[-C] */
extern int OGG_action; /* 0 encode, 1 decode, 2 play */

extern int X_channel;
extern int X_resample;
extern int X_volume;

void setplayback(char *outfilename){
  char buffer[160];
  int temp;
  ALconfig c=ALnewconfig();
  long rates[14]={AL_OUTPUT_RATE, 8000,
		  AL_OUTPUT_RATE, 11025,
		  AL_OUTPUT_RATE, 16000,
		  AL_OUTPUT_RATE, 22050,
		  AL_OUTPUT_RATE, 32000,
		  AL_OUTPUT_RATE, 44100,
		  AL_OUTPUT_RATE, 48000};
		
  if(!c)
    reporterror("Unable to initialize SGI audio");
	
  switch(O_bits){
  case 8:
    temp=AL_SAMPLE_8;
    break;
  default:
    temp=AL_SAMPLE_16;
    O_bits=16;
    break;
  }
	
  if(ALsetwidth(c,temp))
    reporterror("Unable to set SGI sample size.");
	
  switch(O_channels){
  case 1:case 2:
    if(ALsetchannels(c,O_channels))
      reporterror("Unable to set stereo/mono playback");
    break;
  case 4:
    if(ALsetchannels(c,O_channels))
      if(ALsetchannels(c,2))
	reporterror("Unable to set quadraphonic or stereo playback");
      else{
	report("Quadraphonic audio unavailable; using stereo");
	O_channels=2;
	X_channels=2;
      }
    break;
  default:
    if(ALsetchannels(c,O_channels))
      if(ALsetchannels(c,2))
	reporterror("Unable to set stereo playback");
      else{
	report("Weird number of channels; using stereo");
	O_channels=2;
	X_channels=2;
      }
    break;
  }
  
  if(ALsetsampfmt(c,AL_SAMPFMT_TWOSCOMP))
    reporterror("Unable to set SGI integer sample playback");
	

  rates[1]=O_rate;
  
  {
    long i,flag=0;

    /* First try up, then down */

    for(i=0;i<16;i+=2)if(rates[i+1]>=O_rate)break;
    temp=i;
    for(;i<10;i+=2)
      if(ALsetparams(AL_DEFAULT_DEVICE,rates+i,2)==0){

	if(rates[i+1]!=O_rate){
	  sprintf(buffer,"Resampling to %ldHz playback rate",rates[i+1]);
	  report(buffer);
	  X_resample=O_rate=rates[i+1];
	}

	flag=-1;
	break;
      }
    if(!flag)
      for(i=temp-2;i>0;i-=2)
	if(ALsetparams(AL_DEFAULT_DEVICE,rates+i,2)==0){

	  if(rates[i+1]!=O_rate){
	    sprintf(buffer,"Resampling to %ldHz playback rate",rates[i+1]);
	    report(buffer);
	    X_resample=O_rate=rates[i+1];
	  }

	  flag=-1;
	  break;
	}
    if(!flag)
      reporterror("Unable to set any known playback speed");
  }

  /* everything set up as best as we can... */

  if(!(outaudio=ALopenport("OggSQUISH", "w", c)))
    reporterror("Unable to open audio output for playback");

  O_signed=-1;
  O_bigendian=-1;

  ALfreeconfig(c);
}

void closeplayback(void){
  while(ALgetfilled(outaudio))sginap(10);
  ALcloseport(outaudio);
}

void killplayback(void){
  ALcloseport(outaudio);
}

void playback(void *buffer,long bytes,long bytesper){
  if(ALwritesamps(outaudio,buffer,bytes/bytesper))
    reporterror("Unable to write to audio device!");
}

#endif
