/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: D.c
  function: Basic LPC modules
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Sep 10 1996

 ********************************************************************/

#include <math.h>
#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "bitwise.h"
#include "lpc.h"
#include "ilog.h"
#include "D.h"

extern int OM_MACHINE;
extern int __ogg_precision;

static void *LPCcache=NULL;
static long LPCcachebytes=0;
static void *WORKcache=NULL;
static long WORKbytes=0;


void __ogg_cleanupD(void){
  LPCcachebytes=0;
  WORKbytes=0;
}

static void shift_coeffs16(size32 *postshift,size32 *newcoeff,
			   long *coeff,int m){
  int i,j;
  /* This is for residual/predict32 */

  /* We've got an overflow problem, but we do need to operate in fixed point.

     We have two integers to multiply in 32 bits (Most PC users just
     got dragged kicking and screaming into the 32 bit era; it won't
     help that most real machines do at least 64 bit math today.  The
     result we want is always within 32 bits (heck, usually in 16, but
     we can't guarantee it. */

  /* Totally, we always want to right shift 15 bits.  Thus, we
     precalculate beforehand how far to shift before the multiply and
     how much to shift after */

  /* Calculate pre/post shift */
  /* This needs to be based entirely on the coefficient */

  for(i=0;i<m;i++){
    /* max unpreshifted */
    if((j=ilog(labs(coeff[i])))<16){
      postshift[i]=15;
      newcoeff[i]=coeff[i];
    }else{
      postshift[i]=30-j;
      newcoeff[i]=coeff[i]>>(j-15);
    }
  }
}

/* Ogg signal range is up to 32 bits internal; that means that some
signal multiplies are essentially 64 bit math.  Most
platforms/compilers are going to give us 64 bits if we ask, but not
all. */

static void residual32(long *coeff,long *prime,int m,long *data,
			long *residue,long n){

  /* in: coeff[0...m-1] LPC coefficients 
         prime[0...m-1] ... n+m-1]  initial values (allocated size of n+m-1)
	 data[0...n-1] data samples 
    out: residue[0...n-1] residuals from LPC prediction */

  long i,j,y;

  size32 *Iprime;
  size32 *postshift;
  size32 *newcoeff;

  if(LPCcachebytes<sizeof(size32)*(m*3+n)){
    LPCcachebytes=0;
    LPCcache=__ogg_realloc(LPCcache,(m*3+n)*sizeof(size32),
			      OM_MACHINE);
    LPCcachebytes=sizeof(size32)*(n+m*3);
  }

  postshift=((size32 *)LPCcache);
  newcoeff=((size32 *)LPCcache)+m;
  Iprime=((size32 *)LPCcache)+m*2;

  if(prime)
    for(i=0;i<m;i++)
      Iprime[i]=prime[i];
  else
    for(i=0;i<m;i++)
      Iprime[i]=0;

  /* 1.0 == 32768 */
  shift_coeffs16(postshift,newcoeff,coeff,m);

  {
    size32 *ip=Iprime;
    size32 *nc=newcoeff+m-1;
    size32 *ps=postshift+m-1;

    switch(m){
    case 0:
      return;
    case 2:
      for(i=0;i<n;i++){
	y=((*ip)*(*nc))>>(*ps);
	y+=((*(ip+1))*(*(nc-1)))>>(*(ps-1));
	residue[i]=data[i]-y;
	*(ip+2)=data[i];
	ip++;
      }
      break;
    case 4:
      for(i=0;i<n;i++){
	y=((*ip)*(*nc))>>(*ps);
	y+=((*(ip+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(ip+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(ip+3))*(*(nc-3)))>>(*(ps-3));
	residue[i]=data[i]-y;
	*(ip+4)=data[i];
	ip++;
      }
      break;
    case 6:
      for(i=0;i<n;i++){
	y=((*ip)*(*nc))>>(*ps);
	y+=((*(ip+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(ip+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(ip+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(ip+4))*(*(nc-4)))>>(*(ps-4));
	y+=((*(ip+5))*(*(nc-5)))>>(*(ps-5));
	residue[i]=data[i]-y;
	*(ip+6)=data[i];
	ip++;
      }
      break;
    case 8:
      for(i=0;i<n;i++){
	y=((*ip)*(*nc))>>(*ps);
	y+=((*(ip+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(ip+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(ip+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(ip+4))*(*(nc-4)))>>(*(ps-4));
	y+=((*(ip+5))*(*(nc-5)))>>(*(ps-5));
	y+=((*(ip+6))*(*(nc-6)))>>(*(ps-6));
	y+=((*(ip+7))*(*(nc-7)))>>(*(ps-7));
	residue[i]=data[i]-y;
	*(ip+8)=data[i];
	ip++;
      }
      break;
    case 12:
      for(i=0;i<n;i++){
	y=((*ip)*(*nc))>>(*ps);
	y+=((*(ip+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(ip+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(ip+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(ip+4))*(*(nc-4)))>>(*(ps-4));
	y+=((*(ip+5))*(*(nc-5)))>>(*(ps-5));
	y+=((*(ip+6))*(*(nc-6)))>>(*(ps-6));
	y+=((*(ip+7))*(*(nc-7)))>>(*(ps-7));
	y+=((*(ip+8))*(*(nc-8)))>>(*(ps-8));
	y+=((*(ip+9))*(*(nc-9)))>>(*(ps-9));
	y+=((*(ip+10))*(*(nc-10)))>>(*(ps-10));
	y+=((*(ip+11))*(*(nc-11)))>>(*(ps-11));
	residue[i]=data[i]-y;
	*(ip+12)=data[i];
	ip++;
      }
      break;
    case 16:
      for(i=0;i<n;i++){
	y=((*ip)*(*nc))>>(*ps);
	y+=((*(ip+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(ip+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(ip+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(ip+4))*(*(nc-4)))>>(*(ps-4));
	y+=((*(ip+5))*(*(nc-5)))>>(*(ps-5));
	y+=((*(ip+6))*(*(nc-6)))>>(*(ps-6));
	y+=((*(ip+7))*(*(nc-7)))>>(*(ps-7));
	y+=((*(ip+8))*(*(nc-8)))>>(*(ps-8));
	y+=((*(ip+9))*(*(nc-9)))>>(*(ps-9));
	y+=((*(ip+10))*(*(nc-10)))>>(*(ps-10));
	y+=((*(ip+11))*(*(nc-11)))>>(*(ps-11));
	y+=((*(ip+12))*(*(nc-12)))>>(*(ps-12));
	y+=((*(ip+13))*(*(nc-13)))>>(*(ps-13));
	y+=((*(ip+14))*(*(nc-14)))>>(*(ps-14));
	y+=((*(ip+15))*(*(nc-15)))>>(*(ps-15));
	
	residue[i]=data[i]-y;
	
	*(ip+16)=data[i];
	ip++;
      }
      break;
    default:
      for(i=0;i<n;i++){
	y=0;
	for(j=0;j<m;j++)
	  y+=((Iprime[i+j])*newcoeff[m-j-1])>>postshift[m-j-1];
	
	residue[i]=data[i]-y;
	Iprime[i+m]=data[i];
      }

    }
  }
}

#ifdef size64
static void residual64(long *coeff,long *prime,int m,long *data,
			long *residue,long n){

  /* in: coeff[0...m-1] LPC coefficients 
         prime[0...m-1] ... n+m-1]  initial values (allocated size of n+m-1)
	 data[0...n-1] data samples 
    out: residue[0...n-1] residuals from LPC prediction */

  /* Not all platforms have 64 bit math...  gcc can save us in UNIX,
     but perhaps not elsewhere. */

  long i,j;
  size64 y;
  size64 *Iprime;

  if(LPCcachebytes<sizeof(size64)*(n+m)){
    LPCcachebytes=0;
    LPCcache=__ogg_realloc(LPCcache,(n+m)*sizeof(size64),
			    OM_MACHINE);
    LPCcachebytes=sizeof(size64)*(n+m);
  }

  Iprime=(size64 *)LPCcache;

  if(prime)
    for(i=0;i<m;i++)
      Iprime[i]=prime[i];
  else
    for(i=0;i<m;i++)
      Iprime[i]=0;

  /* Don't preshift if we have 64 bits */

  for(i=0;i<n;i++){
    y=0;
    for(j=0;j<m;j++)
      y+=((size64)(Iprime[i+j])*(size64)coeff[m-j-1])>>15;

    residue[i]=data[i]-y;
    Iprime[i+m]=data[i];
  }
}
#endif

static int residualD(long *coeff,long *prime,int m,long *data,
	       long *residue,long n,long bits,long hbits){

  /* in: coeff[0...m-1] LPC coefficients 
         prime[0...m-1] initial values 
	 data[0...n-1] data samples 
    out: residue[0...n-1] residuals from LPC prediction */

  /* Not all platforms have 64 bit math...  gcc can save us in UNIX,
     but perhaps not elsewhere. */

  long i,j;
  double *Fcoeff,*Fprime,y,max=0,*Fres;

  if(LPCcachebytes<sizeof(double)*(m+m+n+n)){
    LPCcachebytes=0;
    LPCcache=__ogg_realloc(LPCcache,(m+m+n+n)*sizeof(double),
			      OM_MACHINE);
    LPCcachebytes=sizeof(double)*(m+m+n+n);
  }

  Fcoeff=(double *)LPCcache;
  Fres=Fcoeff+m;
  Fprime=Fres+n;

  for(i=0;i<m;i++){
    Fcoeff[i]=(double)coeff[i]/32768.;
  }
  if(prime)
    for(i=0;i<m;i++)
      Fprime[i]=(double)prime[i];
  else
    for(i=0;i<m;i++)
      Fprime[i]=0;

  for(i=0;i<n;i++){
    y=0;
    for(j=0;j<m;j++)
      y+=Fprime[i+j]*Fcoeff[m-j-1];

    if(labs(Fres[i]=data[i]-rint(y))>max)max=labs(Fres[i]);
    Fprime[i+m]=data[i];
  }

  if(ilog((long)rint(max))<bits)
    bits-=ilog((long)rint(max));
  else
    if(ilog((long)rint(max))>hbits)
      bits=hbits-ilog((long)rint(max));
    else
      bits=0;
  
  if(bits)
    for(i=0;i<n;i++)
      residue[i]=ldexp(Fres[i],bits);
  else
    for(i=0;i<n;i++)
      residue[i]=Fres[i];

  return(bits);
}

static void predict32(long *coeff,long *prime,int m,long *data,long *residue,
	     long n){
  /* in: coeff[0...m-1] LPC coefficients 
         prime[0...m-1]  initial values 
         residue[0...n-1] residuals from LPC prediction   
    out: data[0...n-1] data samples */
  long i,j,y;
  size32 *postshift;
  size32 *newcoeff;
  size32 *newprime;

  if(LPCcachebytes<sizeof(size32)*(m*3+n)){
    LPCcachebytes=0;
    LPCcache=__ogg_realloc(LPCcache,(m*3+n)*sizeof(size32),
			      OM_MACHINE);
    LPCcachebytes=sizeof(size32)*(n+m*3);
  }

  postshift=(size32 *)LPCcache;
  newcoeff=postshift+m;
  newprime=newcoeff+m;

  if(prime)
    for(i=0;i<m;i++)newprime[i]=prime[i];
  else
    for(i=0;i<m;i++)newprime[i]=0;

  shift_coeffs16(postshift,newcoeff,coeff,m);

  {
    size32 *np=newprime;
    size32 *nc=newcoeff+m-1;
    size32 *ps=postshift+m-1;
    switch(m){
    case 0:
      return;
    case 2:
      for(i=0;i<n;i++){
	y=residue[i]+(((*np)*(*nc))>>(*ps));
	*(np+2)=(data[i]=y+(((*(np+1))*(*(nc-1)))>>(*(ps-1))));
	np++;
      }
      break;
    case 4:
      for(i=0;i<n;i++){
	y=residue[i]+(((*np)*(*nc))>>(*ps));
	y+=((*(np+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(np+2))*(*(nc-2)))>>(*(ps-2));
	*(np+4)=(data[i]=y+(((*(np+3))*(*(nc-3)))>>(*(ps-3))));
	np++;
      }
      break;
    case 6:
      for(i=0;i<n;i++){
	y=residue[i]+(((*np)*(*nc))>>(*ps));
	y+=((*(np+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(np+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(np+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(np+4))*(*(nc-4)))>>(*(ps-4));
	*(np+6)=(data[i]=y+(((*(np+5))*(*(nc-5)))>>(*(ps-5))));
	np++;
      }
      break;
    case 8:
      for(i=0;i<n;i++){
	y=residue[i]+(((*np)*(*nc))>>(*ps));
	y+=((*(np+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(np+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(np+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(np+4))*(*(nc-4)))>>(*(ps-4));
	y+=((*(np+5))*(*(nc-5)))>>(*(ps-5));
	y+=((*(np+6))*(*(nc-6)))>>(*(ps-6));
	*(np+8)=(data[i]=y+(((*(np+7))*(*(nc-7)))>>(*(ps-7))));
	np++;
      }
      break;
    case 12:
      for(i=0;i<n;i++){
	y=residue[i]+(((*np)*(*nc))>>(*ps));
	y+=((*(np+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(np+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(np+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(np+4))*(*(nc-4)))>>(*(ps-4));
	y+=((*(np+5))*(*(nc-5)))>>(*(ps-5));
	y+=((*(np+6))*(*(nc-6)))>>(*(ps-6));
	y+=((*(np+7))*(*(nc-7)))>>(*(ps-7));
	y+=((*(np+8))*(*(nc-8)))>>(*(ps-8));
	y+=((*(np+9))*(*(nc-9)))>>(*(ps-9));
	y+=((*(np+10))*(*(nc-10)))>>(*(ps-10));
	*(np+12)=(data[i]=y+(((*(np+11))*(*(nc-11)))>>(*(ps-11))));
	np++;
      }
      break;
    case 16:
      for(i=0;i<n;i++){
	y=residue[i]+(((*np)*(*nc))>>(*ps));
	y+=((*(np+1))*(*(nc-1)))>>(*(ps-1));
	y+=((*(np+2))*(*(nc-2)))>>(*(ps-2));
	y+=((*(np+3))*(*(nc-3)))>>(*(ps-3));
	y+=((*(np+4))*(*(nc-4)))>>(*(ps-4));
	y+=((*(np+5))*(*(nc-5)))>>(*(ps-5));
	y+=((*(np+6))*(*(nc-6)))>>(*(ps-6));
	y+=((*(np+7))*(*(nc-7)))>>(*(ps-7));
	y+=((*(np+8))*(*(nc-8)))>>(*(ps-8));
	y+=((*(np+9))*(*(nc-9)))>>(*(ps-9));
	y+=((*(np+10))*(*(nc-10)))>>(*(ps-10));
	y+=((*(np+11))*(*(nc-11)))>>(*(ps-11));
	y+=((*(np+12))*(*(nc-12)))>>(*(ps-12));
	y+=((*(np+13))*(*(nc-13)))>>(*(ps-13));
	y+=((*(np+14))*(*(nc-14)))>>(*(ps-14));
	*(np+16)=(data[i]=y+(((*(np+15))*(*(nc-15)))>>(*(ps-15))));
	np++;
      }
      break;
    default:
      
      for(i=0;i<n;i++){
	y=residue[i];
	for(j=0;j<m;j++)
	  y+=(newprime[i+j]*newcoeff[m-j-1])>>postshift[m-j-1];
	
	newprime[i+m]=(data[i]=y);
      }
    }
  }
}

#ifdef size64
static void predict64(long *coeff,long *prime,int m,long *data,long *residue,
	     long n){
  /* in: coeff[0...m-1] LPC coefficients 
         prime[0...m-1] ... n+m-1]  initial values (allocated size of n+m-1)
         residue[0...n-1] residuals from LPC prediction   
    out: data[0...n-1] data samples */
  long i,j;
  size64 y;
  size64 *Iprime;

  if(LPCcachebytes<sizeof(size64)*(n+m)){
    LPCcachebytes=0;
    LPCcache=__ogg_realloc(LPCcache,(n+m)*sizeof(size64),
			    OM_MACHINE);
    LPCcachebytes=sizeof(size64)*(n+m);
  }

  Iprime=(size64 *)LPCcache;

  if(prime)
    for(i=0;i<m;i++)Iprime[i]=prime[i];
  else
    for(i=0;i<m;i++)Iprime[i]=0;

  /* Don't preshift if we have 64 bits */

  for(i=0;i<n;i++){
    y=residue[i];
    for(j=0;j<m;j++)
      y+=((size64)Iprime[i+j]*(size64)coeff[m-j-1])>>15;
  
    Iprime[i+m]=(data[i]=y);
  }
}
#endif 


static void predictD(long *coeff,long *prime,int m,long *data,long *residue,
		     long n,long shift){
  /* in: coeff[0...m-1] LPC coefficients 
         prime[0...m-1] ... n+m-1]  initial values (allocated size of n+m-1)
         residue[0...n-1] residuals from LPC prediction   
    out: data[0...n-1] data samples */
  long i,j,o,p;
  float *Fcoeff,*Fprime,y;

  if(LPCcachebytes<sizeof(double)*(m+m+n+n)){
    LPCcachebytes=0;
    LPCcache=__ogg_realloc(LPCcache,(m+m+n)*sizeof(double),
			      OM_MACHINE);
    LPCcachebytes=sizeof(double)*(m+m+n);
  }

  Fcoeff=(float *)LPCcache;
  Fprime=Fcoeff+m;

  for(i=0;i<m;i++){
    Fcoeff[i]=(float)coeff[i]/32768.;
  }
  if(prime)
    if(shift)
      for(i=0;i<m;i++)
	Fprime[i]=ldexp((float)prime[i],shift);
    else
      for(i=0;i<m;i++)
	Fprime[i]=prime[i];
  else
    for(i=0;i<m;i++)
      Fprime[i]=0;
  
  if(m==4 || m==8 || m==12 || m==16){
    float *ptr1,*ptr2=Fcoeff+m-1;
    for(i=0;i<n;i++){
      ptr1=Fprime+i;
      y=residue[i]+(*ptr1)*(*ptr2);
      y+=(*(ptr1+1))*(*(ptr2-1));
      y+=(*(ptr1+2))*(*(ptr2-2));
      y+=(*(ptr1+3))*(*(ptr2-3));
      if(m>4){
	y+=(*(ptr1+4))*(*(ptr2-4));
	y+=(*(ptr1+5))*(*(ptr2-5));
	y+=(*(ptr1+6))*(*(ptr2-6));
	y+=(*(ptr1+7))*(*(ptr2-7));
	if(m>8){
	  y+=(*(ptr1+8))*(*(ptr2-8));
	  y+=(*(ptr1+9))*(*(ptr2-9));
	  y+=(*(ptr1+10))*(*(ptr2-10));
	  y+=(*(ptr1+11))*(*(ptr2-11));
	  if(m>12){
	    y+=(*(ptr1+12))*(*(ptr2-12));
	    y+=(*(ptr1+13))*(*(ptr2-13));
	    y+=(*(ptr1+14))*(*(ptr2-14));
	    y+=(*(ptr1+15))*(*(ptr2-15));
	  }
	}
      }
      if(shift)
	data[i]=ldexp(Fprime[i+m]=rint(y),-shift);
      else
	data[i]=Fprime[i+m]=rint(y);
    }
  }else{
    for(i=0;i<n;i++){
      y=(float)residue[i];
      o=i;
      p=m;
      for(j=0;j<m;j++)
	y+=Fprime[o++]*Fcoeff[--p];
      
      data[i]=ldexp(Fprime[o]=rint(y),-shift);
    }
  }
}

static void transferD(long *coeff,long amp,int m,long *data,long n){
  /* in: coeff[0...m-1] LPC coefficients 
    out: data[0...n-1] data samples */
  long i,j,o,p;
  float *Fcoeff,*Fprime,y;

  if(LPCcachebytes<sizeof(float)*(m+n+m)){
    LPCcachebytes=0;
    LPCcache=__ogg_realloc(LPCcache,(m+n+m)*sizeof(float),
			      OM_MACHINE);
    LPCcachebytes=sizeof(float)*(m);
  }

  Fcoeff=(float *)LPCcache;
  Fprime=Fcoeff+m;

  for(i=0;i<m;i++){
    Fcoeff[i]=(float)coeff[i]/32768.;
  }
  for(i=0;i<m-1;i++)
    Fprime[i]=0;
  Fprime[m-1]=amp;
  
  for(i=0;i<n;i++){
    y=0;
    o=i;
    p=m;
    for(j=0;j<m;j++)
      y+=Fprime[i+j]*Fcoeff[m-j-1];
      
    data[i]=rint(Fprime[i+m]=y);
  }
}

static void Icoeffs(double *coeff,long *icoeff,int n){
  int count;

  for(count=0;count<n;count++)
    icoeff[count]=(long)-floor(coeff[count]*32768+.5);
}

typedef struct Dl{
  int inpipe;
  int outpipe;
  int var;

  int jump;
} Dl;

/* Dca    inP coeffP aconstV coeffA */

static void __ogga_Dca(module *m){
  Dl *d=(Dl *)(m->state);
  int ip=d->inpipe;
  int op=d->outpipe;
  int v=d->var;
  int poles=__ogg_fixedARG(m,3);

  void *b,*cb;
  int ch,c;
  long *data,*coeff;
  double *work1,*work2;

  b=__ogg_readfromP(ip);
  __ogg_resetP(op);
  if(!b)return;

  c=__ogg_chanMB(b);
  cb=__ogg_allocMB(c,poles);

  if(WORKbytes<(poles*2+1)*sizeof(double)){
    WORKbytes=0;
    WORKcache=__ogg_realloc(WORKcache,(poles*2+1)*sizeof(double),OM_MACHINE);
    WORKbytes=(poles*2+1)*sizeof(double);
  }

  work1=(double *)WORKcache;
  work2=work1+poles+1;

  for(ch=0;ch<c;ch++){
    data=__ogg_ptrMB(b,ch);
    coeff=__ogg_ptrMB(cb,ch);

    memset(work2,0,poles*sizeof(double));
    autocorrelation(__ogg_sizeMB(b),data,poles+1,work1);
    __ogg_writetoV(v,(long)rint(work1[0]/__ogg_sizeMB(b)));
    levinson_durbin(work1,work2,poles);
    Icoeffs(work2,coeff,poles);
  }

  __ogg_appendtoP(ip,b);
  __ogg_appendtoP(op,cb);
}

/* reset is a nop */
  
void __oggc_Dca(module *m){
  Dl *d=(Dl *)(m->state=__ogg_calloc(1,sizeof(Dl),OM_MACHINE));
  d->inpipe=__ogg_fixedARG(m,0);
  d->outpipe=__ogg_fixedARG(m,1);
  d->var=__ogg_fixedARG(m,2);
  m->action=__ogga_Dca;
}

static void __ogga_Dr(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int primeP=__ogg_fixedARG(m,1);
  int coeffP=__ogg_fixedARG(m,2);
  int residueP=__ogg_fixedARG(m,3);
  int bits=__ogg_fixedARG(m,4);
  int hbits=__ogg_fixedARG(m,5);
  int shiftP=__ogg_fixedARG(m,6);

  void *p=__ogg_readfromP(primeP);
  void *b=__ogg_readfromP(dataP);
  void *c=__ogg_readfromP(coeffP);
  int size=__ogg_sizeMB(b);
  int chan=__ogg_chanMB(b),ch;
  int coeffs=__ogg_sizeMB(c);
  void *r=__ogg_allocMB(chan,size);
  void *s=__ogg_allocMB(chan,1);

  if(__ogg_sizeMB(p)<__ogg_sizeMB(c) || __ogg_chanMB(p)!=chan)
    /* ignore the priming data */
    p=NULL;

  for(ch=0;ch<chan;ch++){
    long *prime=__ogg_ptrMB(p,ch);
    long *shiftptr=__ogg_ptrMB(s,ch);
    if(prime)prime+=__ogg_sizeMB(p)-coeffs;
    shiftptr[0]=residualD(__ogg_ptrMB(c,ch),prime,coeffs,__ogg_ptrMB(b,ch),
			  __ogg_ptrMB(r,ch),size,bits,hbits);
  }

  __ogg_appendtoP(primeP,p);
  __ogg_appendtoP(dataP,b);
  __ogg_appendtoP(coeffP,c);
  __ogg_sendtoP(residueP,r);
  __ogg_sendtoP(shiftP,s);
} 

static void __ogga_DrX(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int primeP=__ogg_fixedARG(m,1);
  int coeffP=__ogg_fixedARG(m,2);
  int residueP=__ogg_fixedARG(m,3);

  void *p=__ogg_readfromP(primeP);
  void *b=__ogg_readfromP(dataP);
  void *c=__ogg_readfromP(coeffP);
  int size=__ogg_sizeMB(b);
  int chan=__ogg_chanMB(b),ch;
  int coeffs=__ogg_sizeMB(c);
  void *r=__ogg_allocMB(chan,size);

  if(__ogg_sizeMB(p)<__ogg_sizeMB(c) || __ogg_chanMB(p)!=chan)
    /* ignore the priming data */
    p=NULL;

  for(ch=0;ch<chan;ch++){
    long *prime=__ogg_ptrMB(p,ch);
    if(prime)prime+=__ogg_sizeMB(p)-coeffs;
    if((m->name)[2]=='3')
      residual32(__ogg_ptrMB(c,ch),prime,coeffs,__ogg_ptrMB(b,ch),
		 __ogg_ptrMB(r,ch),size);
#ifdef size64
    else
      residual64(__ogg_ptrMB(c,ch),prime,coeffs,__ogg_ptrMB(b,ch),
		 __ogg_ptrMB(r,ch),size);
#endif
  }

  __ogg_appendtoP(primeP,p);
  __ogg_appendtoP(dataP,b);
  __ogg_appendtoP(coeffP,c);
  __ogg_sendtoP(residueP,r);
} 

void __oggc_Dr(module *m){
  m->action=__ogga_Dr;
}

void __oggc_Dr32(module *m){
  m->action=__ogga_DrX;
}

void __oggc_Dr64(module *m){
#ifdef size64
  m->action=__ogga_DrX;
#else
  __ogg_Cerror(103,0);
#endif
}

/* Dp    dataP primeP coeffP residueP shiftP*/

static void __ogga_Dp(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int primeP=__ogg_fixedARG(m,1);
  int coeffP=__ogg_fixedARG(m,2);
  int residueP=__ogg_fixedARG(m,3);
  int shiftP=__ogg_fixedARG(m,4);
  void *p=__ogg_readfromP(primeP);
  void *b=__ogg_readfromP(residueP);
  void *c=__ogg_readfromP(coeffP);
  void *s=__ogg_readfromP(shiftP);
  int size=__ogg_sizeMB(b);
  int chan=__ogg_chanMB(b),ch;
  int coeffs=__ogg_sizeMB(c);
  void *d=__ogg_allocMB(chan,size);

  if(__ogg_sizeMB(p)<__ogg_sizeMB(c) || __ogg_chanMB(p)!=chan)
    /* ignore the priming data */
    p=NULL;

  for(ch=0;ch<chan;ch++){
    long *prime=__ogg_ptrMB(p,ch);
    int shift;
    if(prime)prime+=__ogg_sizeMB(p)-coeffs;
    if(s)
      shift=__ogg_ptrMB(s,ch)[0];
    else
      shift=0;
    predictD(__ogg_ptrMB(c,ch),prime,coeffs,__ogg_ptrMB(d,ch),
	     __ogg_ptrMB(b,ch),size,shift);
  }
  __ogg_appendtoP(residueP,b);
  __ogg_appendtoP(primeP,p);
  __ogg_appendtoP(coeffP,c);
  __ogg_appendtoP(shiftP,s);
  __ogg_sendtoP(dataP,d);
} 

static void __ogga_DpX(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int primeP=__ogg_fixedARG(m,1);
  int coeffP=__ogg_fixedARG(m,2);
  int residueP=__ogg_fixedARG(m,3);
  void *p=__ogg_readfromP(primeP);
  void *b=__ogg_readfromP(residueP);
  void *c=__ogg_readfromP(coeffP);
  int size=__ogg_sizeMB(b);
  int chan=__ogg_chanMB(b),ch;
  int coeffs=__ogg_sizeMB(c);
  void *d=__ogg_allocMB(chan,size);

  if(__ogg_sizeMB(p)<__ogg_sizeMB(c) || __ogg_chanMB(p)!=chan)
    /* ignore the priming data */
    p=NULL;

  for(ch=0;ch<chan;ch++){
    long *prime=__ogg_ptrMB(p,ch);
    if(prime)prime+=__ogg_sizeMB(p)-coeffs;
    if((m->name)[2]=='3')
      predict32(__ogg_ptrMB(c,ch),prime,coeffs,__ogg_ptrMB(d,ch),
		__ogg_ptrMB(b,ch),size);
#ifdef size64
    else
      predict64(__ogg_ptrMB(c,ch),prime,coeffs,__ogg_ptrMB(d,ch),
		__ogg_ptrMB(b,ch),size);
#endif
  }
  __ogg_appendtoP(residueP,b);
  __ogg_appendtoP(primeP,p);
  __ogg_appendtoP(coeffP,c);
  __ogg_sendtoP(dataP,d);
} 

void __oggc_Dp(module *m){
  m->action=__ogga_Dp;
}

void __oggc_Dp32(module *m){
  m->action=__ogga_DpX;
}

void __oggc_Dp64(module *m){
#ifdef size64
  m->action=__ogga_DpX;
#else
  __ogg_Cerror(103,0);
#endif
}

static void __ogga_Ddf(module *m){
  int inP=__ogg_fixedARG(m,0);
  int primeP=__ogg_fixedARG(m,1);
  int outP=__ogg_fixedARG(m,2);

  void *b=__ogg_copyfromP(inP);
  void *p=__ogg_readfromP(primeP);
  int size=__ogg_sizeMB(b);
  int chan=__ogg_chanMB(b),ch;

  for(ch=0;ch<chan;ch++){
    long *vector=__ogg_ptrMB(b,ch);
    long *prime=__ogg_ptrMB(p,ch);
    int n=size;

    while(--n)vector[n]-=vector[n-1];
    if(prime)
      vector[0]-=prime[__ogg_sizeMB(p)-1];
  }

  __ogg_appendtoP(primeP,p);
  __ogg_sendtoP(outP,b);
}

static void __ogga_Ddi(module *m){
  int inP=__ogg_fixedARG(m,0);
  int primeP=__ogg_fixedARG(m,1);
  int outP=__ogg_fixedARG(m,2);

  void *b=__ogg_copyfromP(inP);
  void *p=__ogg_readfromP(primeP);
  int size=__ogg_sizeMB(b);
  int chan=__ogg_chanMB(b),ch;

  for(ch=0;ch<chan;ch++){
    long *vector=__ogg_ptrMB(b,ch);
    long *prime=__ogg_ptrMB(p,ch);
    int n=0;

    if(prime)
      vector[0]+=prime[__ogg_sizeMB(p)-1];
    while(++n<size)vector[n]+=vector[n-1];
  }

  __ogg_appendtoP(primeP,p);
  __ogg_sendtoP(outP,b);
}

void __oggc_Ddf(module *m){
  m->action=__ogga_Ddf;
}

void __oggc_Ddi(module *m){
  m->action=__ogga_Ddi;
}

/* Dtrans outP coeffP ampA lenA */

static void __ogga_Dtrans(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int coeffP=__ogg_fixedARG(m,1);
  long amp=__ogg_fixedARG(m,2);
  long size=__ogg_fixedARG(m,3);

  void *c=__ogg_readfromP(coeffP);
  int chan=__ogg_chanMB(c),ch;
  int coeffs=__ogg_sizeMB(c);
  void *b=__ogg_allocMB(chan,size);

  for(ch=0;ch<chan;ch++)
    transferD(__ogg_ptrMB(c,ch),amp,coeffs,__ogg_ptrMB(b,ch),size);

  __ogg_sendtoP(dataP,b);
  __ogg_sendtoP(coeffP,c);
} 

void __oggc_Dtrans(module *m){
  m->action=__ogga_Dtrans;
}

