/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1997             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: E.c
  function: Vector and small matrix transforms
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Dec 16 1997

 ********************************************************************/

#include <math.h>
#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "quantize.h"
#include "E.h"
#include "ilog.h"

extern int OM_MACHINE;

static void __ogga_Es(module *m){
  int ip=__ogg_fixedARG(m,0);
  int ap=__ogg_fixedARG(m,1);
  int bp=__ogg_fixedARG(m,2);
  long split=__ogg_fixedARG(m,3);
  void *b=__ogg_readnfromP(ip,split);
  __ogg_sendtoP(bp,__ogg_readfromP(ip));
  __ogg_sendtoP(ap,b);
}

void __oggc_Es(module *m){
  m->action=__ogga_Es;
}

static void __ogga_Et(module *m){
  int ip=__ogg_fixedARG(m,0);
  long from=__ogg_fixedARG(m,1);
  long to=__ogg_fixedARG(m,2);
  void *null1=NULL,*b;
  long size;

  if(from)null1=__ogg_readnfromP(ip,from);
    __ogg_freeMB(null1);

  b=__ogg_readfromP(ip);
  size=__ogg_sizeMB(b);
  if(to>from){
    if(to-from<size)
      __ogg_cutMB(b,to-from);
    __ogg_sendtoP(ip,b);
    return;
  }
  __ogg_freeMB(b);
}

void __oggc_Et(module *m){
  m->action=__ogga_Et;
}

static void __ogga_Ez(module *m){
  int op=__ogg_fixedARG(m,0);
  long size=__ogg_fixedARG(m,1);
  long chan=__ogg_fixedARG(m,2);
  long val=__ogg_fixedARG(m,3);
  void *b=__ogg_allocMB(chan,size);
  if(val)__ogg_setMB(b,val);
  __ogg_sendtoP(op,b);
}

void __oggc_Ez(module *m){
  m->action=__ogga_Ez;
}

static void __ogga_Epad(module *m){
  int inP=__ogg_fixedARG(m,0);
  long left=__ogg_fixedARG(m,1);
  long right=__ogg_fixedARG(m,2);
  void *b=__ogg_readfromP(inP);
  long chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b);

  void *out=__ogg_allocMB(chan,size+left+right);

  for(ch=0;ch<chan;ch++){
    long *inptr=__ogg_ptrMB(b,ch);
    long *outptr=__ogg_ptrMB(out,ch);

    memcpy(outptr+left,inptr,size*sizeof(long));
  }
  __ogg_freeMB(b);
  __ogg_sendtoP(inP,out);
}

void __oggc_Epad(module *m){
  m->action=__ogga_Epad;
}

static void __ogga_Ezo(module *m){
  int p=__ogg_fixedARG(m,0);
  long from=__ogg_fixedARG(m,1);
  long to=__ogg_fixedARG(m,2);
  long val=__ogg_fixedARG(m,3);
  void *b=__ogg_readfromP(p);
  long chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b),i;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    for(i=from;i<size && i<=to;i++)
      ptr[i]=val;
  }
  __ogg_appendtoP(p,b);
}

void __oggc_Ezo(module *m){
  m->action=__ogga_Ezo;
}

static void __ogga_Etr(module *m){ /*PAA*/
  int p=__ogg_fixedARG(m,0);
  long forevery=__ogg_fixedARG(m,1);
  long samplen=__ogg_fixedARG(m,2);
  void *b=__ogg_readfromP(p);
  long chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b),i,j,k;
  void *o=__ogg_allocMB(chan,size/forevery*samplen);

  if(forevery>samplen){

    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      long *out=__ogg_ptrMB(o,ch);
      for(i=0,j=0;i+forevery<=size;i+=forevery)
	for(k=0;k<samplen;k++)
	  out[j++]=ptr[i+k];
    }

  }else{

    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      long *out=__ogg_ptrMB(o,ch);
      for(i=0,j=0;i+forevery<=size;i+=forevery){
	for(k=0;k<forevery;k++)
	  out[j++]=ptr[i+k];
        for(;k<samplen;k++)
	  out[j++]=0;
      }

    }
    
  }
  __ogg_freeMB(b);
  __ogg_appendtoP(p,o);
}

void __oggc_Etr(module *m){
  m->action=__ogga_Etr;
}

static void __ogga_Ec(module *m){
  int ap=__ogg_fixedARG(m,0);
  int bp=__ogg_fixedARG(m,1);
  int op=__ogg_fixedARG(m,2);
  void *temp=__ogg_readfromP(bp);
  __ogg_sendtoP(op,__ogg_readfromP(ap));
  __ogg_appendtoP(op,temp);
}

void __oggc_Ec(module *m){
  m->action=__ogga_Ec;
}

static void __ogga_Eq(module *m){
  int p=__ogg_fixedARG(m,0);
  int bits=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(p);
  int chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b);

  if(bits<1 || bits>31)return;
  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    switch(bits){
    case 1:
      __ogg_msb_quantize(ptr,size);
      break;
    default:
      __ogg_msb_quantizeN(ptr,size,bits);
      break;
    }
  }
  __ogg_appendtoP(p,b);
}

void __oggc_Eq(module *m){
  m->action=__ogga_Eq;
}

static void __ogga_Eqi(module *m){
  int p=__ogg_fixedARG(m,0);
  int bits=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(p);
  int chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b);

  if(bits<1 || bits>31)return;
  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    switch(bits){
    case 1:
      __ogg_msb_unquantize(ptr,size);
      break;
    default:
      __ogg_msb_unquantizeN(ptr,size,bits);
      break;
    }
  }
  __ogg_appendtoP(p,b);
}

void __oggc_Eqi(module *m){
  m->action=__ogga_Eqi;
}

static void __ogga_El(module *m){
  int inP=__ogg_fixedARG(m,0);
  int shiftP=__ogg_fixedARG(m,1);
  long interval=__ogg_fixedARG(m,2);
  long maxval=__ogg_fixedARG(m,3);

  void *buf=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(buf),ch;
  long size=__ogg_sizeMB(buf),i,j;
  long size2=(size+interval-1)/interval;
  void *shiftbuf=__ogg_allocMB(chan,size2);

  if (size==0)return;

  for(ch=0;ch<chan;ch++){
    long *shiftptr=__ogg_ptrMB(shiftbuf,ch);
    long *ptr=__ogg_ptrMB(buf,ch);
    long begin=0; 
    long end=(interval>size?size:interval);

    for(i=0;i<size2;i++){
      int max=0;
      long shift=0;

      for(j=begin;j<end;j++)
	if(labs(ptr[j])>max)max=labs(ptr[j]);

      if(max>maxval){
	shift=max-maxval;

	for(j=begin;j<end;j++)
	  if(ptr[j]<-shift)
	    ptr[j]+=shift;
	  else
	    if(ptr[j]>shift)
	      ptr[j]-=shift;
	    else
	      ptr[j]=0;
      }
      shiftptr[i]=shift;

      begin=end;
      end=(end+interval>size?size:end+interval);
    }
  }
  __ogg_sendtoP(inP,buf);
  __ogg_sendtoP(shiftP,shiftbuf);
}

void __oggc_El(module *m){
  m->action=__ogga_El;
}

static void __ogga_Eli(module *m){
  int inP=__ogg_fixedARG(m,0);
  int shiftP=__ogg_fixedARG(m,1);

  void *buf=__ogg_readfromP(inP);
  void *shiftbuf=__ogg_readfromP(shiftP);

  int chan=__ogg_chanMB(buf),ch;
  long size=__ogg_sizeMB(buf),i,j;
  int chan2=__ogg_chanMB(shiftbuf);
  long size2=__ogg_sizeMB(shiftbuf);

  if(size2==0)return;
  if(chan<chan2)chan=chan2;

  for(ch=0;ch<chan;ch++){
    long *shiftptr=__ogg_ptrMB(shiftbuf,ch);
    long *ptr=__ogg_ptrMB(buf,ch);

    long begin=0; 
    long end=size/size2;

    for(i=0;i<size2;i++){
      for(j=begin;j<end;j++)
	if(ptr[j]<0)
	  ptr[j]-=shiftptr[i];
	else
	  if(ptr[j]>0)
	    ptr[j]+=shiftptr[i];
      
      
      begin=end;
      end=size*(i+2)/size2;
    }
  }
  __ogg_sendtoP(inP,buf);
  __ogg_sendtoP(shiftP,shiftbuf);
}

void __oggc_Eli(module *m){
  m->action=__ogga_Eli;
}

static void __ogga_Edi(module *m){
  int in=__ogg_fixedARG(m,0);
  int out=__ogg_fixedARG(m,1);
  int drift=__ogg_fixedARG(m,2);
  void *b=__ogg_readfromP(in);
  void *d=__ogg_readfromP(drift);
  int chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b),i;
  void *o=__ogg_allocMB(chan,size);

  if(b && d){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      long *optr=__ogg_ptrMB(o,ch);
      long drift=__ogg_ptrMB(d,ch)[0];
      long next;

      optr[0]=drift-ptr[size-1];
      next=2;

      while(next<=size){
	for(i=next-1;i>2;i-=2){
	  optr[i]=optr[i>>1];
	  optr[i-1]=(optr[i]+optr[(i>>1)-1])>>1;
	}
	optr[1]=optr[0];
	optr[0]>>=1;

	next<<=1;
      }

      for(i=0;i<size;i++)
	optr[i]+=ptr[i];

    }
  }
  __ogg_appendtoP(in,b);
  __ogg_appendtoP(drift,d);
  __ogg_sendtoP(out,o);
}

void __oggc_Edi(module *m){
  m->action=__ogga_Edi;
}

static void __ogga_Eaa(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);

    for(i=0;i<size;i++)
      outptr[i]+=inptr[i];
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_Eaa(module *m){
  m->action=__ogga_Eaa;
}

/* Vector element multiplex; take each n scalar elements of m values and
   multiplex them into singular scalar elements of n*m values.
   Lookup table based */

static void __ogga_Evmx(module *m){

  int inP=__ogg_fixedARG(m,0);
  int multP=__ogg_fixedARG(m,1);

  int range=__ogg_fixedARG(m,2);
  int nvals=__ogg_fixedARG(m,3);

  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(in),ch;
  long size=__ogg_sizeMB(in),i,k;

  long size2=size/nvals;
  void *out=__ogg_allocMB(chan,size2);

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);
    long j=0;

    /* currently dangerous; we don't guard bounds */

    switch(nvals){
    case 2:
      for(i=0;i+2<=size;i+=2)
	outptr[j++]=inptr[i+1]*range+inptr[i];
      break;
    case 3:
      for(i=0;i+3<=size;i+=3)
	outptr[j++]=(inptr[i+2]*range+inptr[i+1])*range+inptr[i];
      break;
    case 4:
      for(i=0;i+4<=size;i+=4)
	outptr[j++]=((inptr[i+3]*range+inptr[i+2])*range+inptr[i+1])*range+
	  inptr[i];
      break;
    case 5:
      for(i=0;i+5<=size;i+=5)
	outptr[j++]=(((inptr[i+4]*range+inptr[i+3])*range+inptr[i+2])*range+
	  inptr[i+1])*range+inptr[i];
      break;
    default:
      for(i=0;i+nvals<=size;i+=nvals){
	outptr[j]=inptr[i+nvals-1];
	for(k=2;k<=nvals;k++)
	  outptr[j]=(outptr[j]*range)+inptr[i+nvals-k];
      }
    }
  }
  __ogg_sendtoP(inP,in);
  __ogg_sendtoP(multP,out);
}

void __oggc_Evmx(module *m){
  m->action=__ogga_Evmx;
}

typedef struct Evmxi {
  int *lookup;
  int range;
  int nvals;
  int shift;
} Evmxi;

static void init_Evmxi(Evmxi *e,int range,int nvals){

  /* to save time and waste space, we take the next largest power of
     two from nvals */
  int size,keys=1,i;
  e->shift=ilog((e->nvals=nvals)-1);
  e->range=range;
  for(i=0;i<nvals;i++)keys*=range;

  size=keys*(1<<e->shift);

  __ogg_free(e->lookup);
  e->lookup=__ogg_calloc(size,sizeof(int),OM_MACHINE);

  {
    long k;
    
    for(i=0;i<keys;i++){
      long temp=i;
      long index=i<<e->shift;
     
      for(k=0;k<nvals;k++){
	e->lookup[index+k]=temp%range;
	temp/=range;
      }
    }
  }
} 

static void __ogga_Evmxi(module *m){
  Evmxi *e=(Evmxi *)m->state;
  int inP=__ogg_fixedARG(m,0);
  int expP=__ogg_fixedARG(m,1);
  long range=__ogg_fixedARG(m,2);
  long nvals=__ogg_fixedARG(m,3);

  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(in),ch;
  long size=__ogg_sizeMB(in),i,k;

  long size2=size*nvals;
  void *out=__ogg_allocMB(chan,size2);

  if (e->lookup==NULL || 
      e->range!=range ||
      e->nvals!=nvals)init_Evmxi(e,range,nvals);

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);
    long j=0,index;

    for(i=0;i<size;i++){
      index=inptr[i]<<e->shift;
      for(k=0;k<nvals;k++)
	outptr[j+k]=e->lookup[index+k];
      j+=nvals;
    }
  }
  __ogg_sendtoP(inP,in);
  __ogg_sendtoP(expP,out);
}

void __oggc_Evmxi(module *m){
  m->state=__ogg_calloc(1,sizeof(Evmxi),OM_MACHINE);  
  m->action=__ogga_Evmxi;
}

/* in v99, this doesn't exist; the vector version has repeating behavior */

static void __ogga_Easa(module *m){
  int outP=__ogg_fixedARG(m,0);
  long val=__ogg_fixedARG(m,1);
  void *out=__ogg_readfromP(outP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i<size;i++)
      outptr[i]+=val;
  }
  __ogg_appendtoP(outP,out);
}

void __oggc_Easa(module *m){
  m->action=__ogga_Easa;
}

static void __ogga_Efd(module *m){
  int outP=__ogg_fixedARG(m,0);
  long poweroftwo=__ogg_fixedARG(m,1),i;
  void *out=__ogg_readfromP(outP);
  int chan=__ogg_chanMB(out),ch;

  for(i=0;i<poweroftwo;i++){
    long size=__ogg_sizeMB(out),j;

    if(size&1)break;

    for(ch=0;ch<chan;ch++){
      long *b=__ogg_ptrMB(out,ch);

      b[0]+=b[size-1];
      b[size/2-1]*=2;

      for(j=1;j<size/4;j++){
	b[j*2-1]+=b[size-j*2-1];
	b[j*2]-=b[size-j*2];
      }
    }

    __ogg_cutMB(out,size/2);

  }

  __ogg_appendtoP(outP,out);
}

void __oggc_Efd(module *m){
  m->action=__ogga_Efd;
}


static void __ogga_Eas(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);

    for(i=0;i<size;i++)
      outptr[i]-=inptr[i];
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_Eas(module *m){
  m->action=__ogga_Eas;
}

static void __ogga_Eam(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);

    for(i=0;i<size;i++)
      outptr[i]*=inptr[i];
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_Eam(module *m){
  m->action=__ogga_Eam;
}

static void __ogga_Ecm(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);

    outptr[0]*=inptr[0];
    outptr[size-1]*=inptr[size-1];

    for(i=1;i<size-1;i+=2){
      long a=outptr[i];
      long b=outptr[i+1];
      long c=inptr[i];
      long d=inptr[i+1];
      outptr[i]=a*c-b*d;
      outptr[i+1]=b*c+a*d;
    }
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_Ecm(module *m){
  m->action=__ogga_Ecm;
}

/* transpose, fft style packing */
static void __ogga_Ectr(module *m){
  int inP=__ogg_fixedARG(m,0);
  void *in=__ogg_readfromP(inP);

  int chan=__ogg_chanMB(in),ch;
  long size=__ogg_sizeMB(in),i;

  for(ch=0;ch<chan;ch++){
    long *inptr=__ogg_ptrMB(in,ch);

    long temp=inptr[0];
    inptr[0]=inptr[size-1];
    inptr[size-1]=temp;

    for(i=1;i<size/2;i+=2){
      long a=inptr[i];
      long b=inptr[i+1];
      inptr[i]=inptr[size-i-2];
      inptr[i+1]=inptr[size-i-1];
      inptr[size-i-2]=a;
      inptr[size-i-1]=b;
    }
  }
  __ogg_appendtoP(inP,in);
}

void __oggc_Ectr(module *m){
  m->action=__ogga_Ectr;
}

static void __ogga_EcdFs(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  float scale=__ogg_fixedARG(m,2);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);

    outptr[0]=rint(outptr[0]*scale/inptr[0]);
    outptr[size-1]=rint(outptr[size-1]*scale/inptr[size-1]);

    for(i=1;i<size-1;i+=2){
      float Ar=outptr[i];
      float Ai=outptr[i+1];
      float Br=inptr[i];
      float Bi=inptr[i+1];

      float r=Ar*Br+Ai*Bi;
      float j=Ai*Br-Ar*Bi;
      float d=scale/(Br*Br+Bi*Bi);
      
      
      /*
	A/B = (Ar+iAi) / (Br+iBi) = 

	(Ar+iAi)*(Br-iBi)
	----------------- =
        (Br+iBi)*(Br-iBi)

	Ar*Br+Ai*Bi + i(Ai*Br - Ar*Bi)
        ------------------------------
	Br*Br+Bi*Bi

	*/
      if(d!=0){
	outptr[i]=rint(r*d);
	outptr[i+1]=rint(j*d);
      }else{
	outptr[i]=0;
	outptr[i+1]=0;
      }
    }
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_EcdFs(module *m){
  m->action=__ogga_EcdFs;
}

static void __ogga_Ecmag(module *m){
  int inP=__ogg_fixedARG(m,0);

  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(in),ch;
  long size=__ogg_sizeMB(in),i;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(in,ch);

    /* 0 is pure real */

    for(i=1;i<size-1;i+=2){
      float a=ptr[i];
      float b=ptr[i+1];
      ptr[(i>>1)+1]=sqrt(a*a+b*b);
    }

    ptr[size>>1]=labs(ptr[size-1]);
    ptr[0]=labs(ptr[0]);

  }

  __ogg_cutMB(in,(size>>1)+1);
  __ogg_appendtoP(inP,in);
}

void __oggc_Ecmag(module *m){
  m->action=__ogga_Ecmag;
}

static void __ogga_Ecms(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  long scale=__ogg_fixedARG(m,2);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);
    float scalef=1./scale;

    outptr[0]=rint(outptr[0]*scalef*inptr[0]);
    outptr[size-1]=rint(scalef*outptr[size-1]*inptr[size-1]);

    for(i=1;i<size-1;i+=2){
      float a=outptr[i];
      float b=outptr[i+1];
      float c=inptr[i];
      float d=inptr[i+1];
      outptr[i]=rint((a*c-b*d)*scalef);
      outptr[i+1]=rint((b*c+a*d)*scalef);
    }
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_Ecms(module *m){
  m->action=__ogga_Ecms;
}

static void __ogga_Efu(module *m){
  int inP=__ogg_fixedARG(m,0);
  long poweroftwo=__ogg_fixedARG(m,1),i;
  void *in=__ogg_readfromP(inP);
  
  int chan=__ogg_chanMB(in),ch;
  long size=__ogg_sizeMB(in),j;

  void *out=__ogg_allocMB(chan,size<<poweroftwo);

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);
    for(i=0;i<size;i++)outptr[i]=inptr[i];
  }

  for(j=0;j<poweroftwo;j++){
    long size2=size<<1;
    for(ch=0;ch<chan;ch++){
      long *outptr=__ogg_ptrMB(out,ch);
      outptr[size2-1]=outptr[0];
      outptr[size]=0;
      
      for(i=1;i<size-1;i+=2){
	outptr[size2-i-2]=outptr[i];
	outptr[size2-i-1]=-outptr[i+1];
      }
    }
    size=size2;
  }
  __ogg_appendtoP(inP,out);
  __ogg_freeMB(in);
}

void __oggc_Efu(module *m){
  m->action=__ogga_Efu;
}

static void __ogga_Ean(module *m){
  int outP=__ogg_fixedARG(m,0);
  void *out=__ogg_readfromP(outP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i<size;i++)
      outptr[i]= -outptr[i];
  }
  __ogg_appendtoP(outP,out);
}

void __oggc_Ean(module *m){
  m->action=__ogga_Ean;
}

static void __ogga_Eab(module *m){
  int outP=__ogg_fixedARG(m,0);
  void *out=__ogg_readfromP(outP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i<size;i++)
      outptr[i]=labs(outptr[i]);
  }
  __ogg_appendtoP(outP,out);
}

void __oggc_Eab(module *m){
  m->action=__ogga_Eab;
}

static void __ogga_Easm(module *m){
  int outP=__ogg_fixedARG(m,0);
  int n=__ogg_fixedARG(m,1);
  int d=__ogg_fixedARG(m,2);

  void *out=__ogg_readfromP(outP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i<size;i++)
      outptr[i]= outptr[i]*n/d;
  }
  __ogg_appendtoP(outP,out);
}

void __oggc_Easm(module *m){
  m->action=__ogga_Easm;
}


static void __ogga_EasmF(module *m){
  int outP=__ogg_fixedARG(m,0);
  int n=__ogg_fixedARG(m,1);
  int d=__ogg_fixedARG(m,2);
  float fm=(float)n/(float)d;

  void *out=__ogg_readfromP(outP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i<size;i++)
      outptr[i]*=fm;
  }
  __ogg_appendtoP(outP,out);
}

void __oggc_EasmF(module *m){
  m->action=__ogga_EasmF;
}

/* Split a single channel vector out of a multichannel vector pack */

static void __ogga_Ecs(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int channelp=__ogg_fixedARG(m,1);
  int outP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  int chan=__ogg_chanMB(data);
  long size=__ogg_sizeMB(data);
  void *out=__ogg_allocMB(1,size);

  if(channelp>=0 && channelp<chan){
    long *inptr=__ogg_ptrMB(data,channelp);
    long *outptr=__ogg_ptrMB(out,0);
    memcpy(outptr,inptr,size*sizeof(long));
  }
  __ogg_appendtoP(dataP,data);
  __ogg_sendtoP(outP,out);
}

void __oggc_Ecs(module *m){
  m->action=__ogga_Ecs;
}

/* replace a channel in an input bundle with data from an input */

static void __ogga_Ecr(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int channelp=__ogg_fixedARG(m,1);
  int inP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  void *in=__ogg_readfromP(inP);

  int chan=__ogg_chanMB(data);
  long size=__ogg_sizeMB(data);
  long size2=__ogg_sizeMB(in);
  if(size2<size)size=size2;

  if(channelp>=0 || channelp<chan){
    long *dataptr=__ogg_ptrMB(data,channelp);
    long *inptr=__ogg_ptrMB(in,0);
    memcpy(dataptr,inptr,size*sizeof(long));
  }
  __ogg_appendtoP(dataP,data);
  __ogg_appendtoP(inP,in);
}

void __oggc_Ecr(module *m){
  m->action=__ogga_Ecr;
}

static void __ogga_Eav(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int ch=__ogg_fixedARG(m,1);
  int outV=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);

  int chan=__ogg_chanMB(data);
  long size=__ogg_sizeMB(data),i;

  double acc=0;
  
  if(ch<0){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(data,ch);
      for(i=0;i<size;i++)
	acc+=labs(ptr[i]);
    }
    if(size+chan>0)
      acc/=(size*chan);
  }else{
    if(ch>=0 && ch<chan){
      long *ptr=__ogg_ptrMB(data,ch);
      for(i=0;i<size;i++)
	acc+=labs(ptr[i]);
    }
    if(size>0)
      acc/=size;
  }
  
  __ogg_appendtoP(dataP,data);
  __ogg_writetoV(outV,(long)rint(acc));
}

void __oggc_Eav(module *m){
  m->action=__ogga_Eav;
}

static void __ogga_Emax(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int outP=__ogg_fixedARG(m,1);

  void *data=__ogg_readfromP(dataP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;
  void *out=__ogg_allocMB(chan,1);

  for(ch=0;ch<chan;ch++){
    long max=0;
    long *ptr=__ogg_ptrMB(data,ch);
    long *outptr=__ogg_ptrMB(out,ch);
    for(i=0;i<size;i++)
      if(labs(ptr[i])>max)max=labs(ptr[i]);
    outptr[0]=max;
  }
  
  __ogg_sendtoP(dataP,data);
  __ogg_sendtoP(outP,out);
}

void __oggc_Emax(module *m){
  m->action=__ogga_Emax;
}

static void __ogga_Esum(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int outP=__ogg_fixedARG(m,1);

  void *data=__ogg_readfromP(dataP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;
  void *out=__ogg_allocMB(chan,1);

  for(ch=0;ch<chan;ch++){
    long acc=0;
    long *ptr=__ogg_ptrMB(data,ch);
    long *outptr=__ogg_ptrMB(out,ch);
    for(i=0;i<size;i++)acc+=labs(ptr[i]);
    outptr[0]=acc;
  }
  
  __ogg_sendtoP(dataP,data);
  __ogg_sendtoP(outP,out);
}

void __oggc_Esum(module *m){
  m->action=__ogga_Esum;
}

static void __ogga_Eqlog(module *m){
  int dataP=__ogg_fixedARG(m,0);
  long vals=__ogg_fixedARG(m,1);
  long baseP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  void *base=__ogg_readfromP(baseP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;

  long scale=vals/2;
  long odd=vals%2;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(data,ch);
    long *baseptr=__ogg_ptrMB(base,ch);
    float logd;
    if(baseptr[0]==0)
      logd=0;
    else
      logd=(float)scale/log(baseptr[0]);


    if(odd)
      for(i=0;i<size;i++)
	if(ptr[i]<0){
	  ptr[i]=rint(-log(-ptr[i])*logd);
	  if(ptr[i]<-scale)ptr[i]=-scale;
	}else{
	  ptr[i]=rint(log(ptr[i])*logd);
	  if(ptr[i]>=scale)ptr[i]=scale-1;
	}
    else
      for(i=0;i<size;i++)
	if(ptr[i]<0){
	  ptr[i]=rint(-log(-ptr[i])*logd-.5);
	  if(ptr[i]<-scale)ptr[i]=-scale;
	}else{
	  ptr[i]=rint(log(ptr[i])*logd-.5);
	  if(ptr[i]>=scale)ptr[i]=scale-1;
	}
  }
  
  __ogg_sendtoP(dataP,data);
  __ogg_sendtoP(baseP,base);
}

void __oggc_Eqlog(module *m){
  m->action=__ogga_Eqlog;
}

static void __ogga_Eexp(module *m){
  int dataP=__ogg_fixedARG(m,0);
  long num=__ogg_fixedARG(m,1);
  long denom=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;

  float power=(float)num/denom;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(data,ch);

    for(i=0;i<size;i++)
      if(ptr[i]<0)
	ptr[i]=rint(-pow(-ptr[i],power));
      else
	ptr[i]=rint(pow(ptr[i],power));

  }
  
  __ogg_sendtoP(dataP,data);
}

void __oggc_Eexp(module *m){
  m->action=__ogga_Eexp;
}

static void __ogga_Eqlogi(module *m){
  int dataP=__ogg_fixedARG(m,0);
  long vals=__ogg_fixedARG(m,1);
  long baseP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  void *base=__ogg_readfromP(baseP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;

  float scale=2./vals;
  long odd=vals%2;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(data,ch);
    long baseval=__ogg_ptrMB(base,ch)[0];
    
    if(odd)
      for(i=0;i<size;i++)
	if(ptr[i]<0)
	  ptr[i]=rint(-pow(baseval,-ptr[i]*scale));
	else
	  ptr[i]=rint(pow(baseval,ptr[i]*scale));
    else
      for(i=0;i<size;i++)
	if(ptr[i]<0)
	  ptr[i]=rint(-pow(baseval,-(ptr[i]+.5)*scale));
	else
	  ptr[i]=rint(pow(baseval,(ptr[i]+.5)*scale));
  }
  
  __ogg_sendtoP(dataP,data);
  __ogg_sendtoP(baseP,base);
}

void __oggc_Eqlogi(module *m){
  m->action=__ogga_Eqlogi;
}

static void __ogga_Eqlin(module *m){
  int dataP=__ogg_fixedARG(m,0);
  long vals=__ogg_fixedARG(m,1);
  long baseP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  void *base=__ogg_readfromP(baseP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;

  long scale=vals/2;
  long odd=vals%2;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(data,ch);
    long *baseptr=__ogg_ptrMB(base,ch);
    float mult;
    if(baseptr[0]==0)
      mult=0;
    else
      mult=(float)scale/baseptr[0];

    if(odd)
      for(i=0;i<size;i++){
	ptr[i]=rint(ptr[i]*mult);
	if(ptr[i]<-scale)ptr[i]=-scale;
	if(ptr[i]>=scale)ptr[i]=scale-1;
      }
    else
      for(i=0;i<size;i++){
	ptr[i]=rint(ptr[i]*mult-.5);
	if(ptr[i]<-scale)ptr[i]=-scale;
	if(ptr[i]>=scale)ptr[i]=scale-1;
      }
  }
  
  __ogg_sendtoP(dataP,data);
  __ogg_sendtoP(baseP,base);
}

void __oggc_Eqlin(module *m){
  m->action=__ogga_Eqlin;
}

static void __ogga_Eqlini(module *m){
  int dataP=__ogg_fixedARG(m,0);
  long vals=__ogg_fixedARG(m,1);
  long baseP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  void *base=__ogg_readfromP(baseP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;

  long scale=vals/2;
  long odd=vals%2;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(data,ch);
    long baseval=__ogg_ptrMB(base,ch)[0];
    float mult=(float)baseval/scale;
    
    if(odd)
      for(i=0;i<size;i++)
	ptr[i]=rint(ptr[i]*mult);
    else
      for(i=0;i<size;i++)
	ptr[i]=rint((ptr[i]+.5)*mult);
  }
  
  __ogg_sendtoP(dataP,data);
  __ogg_sendtoP(baseP,base);
}

void __oggc_Eqlini(module *m){
  m->action=__ogga_Eqlini;
}

static void __ogga_Est(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int resultP=__ogg_fixedARG(m,1);

  void *data=__ogg_readfromP(dataP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;

  void *result=__ogg_allocMB(1,chan>>1);
  long *resultptr=__ogg_ptrMB(result,0);

  for(ch=0;ch+1<chan;ch+=2){
    long *ptrA=__ogg_ptrMB(data,ch);
    long *ptrB=__ogg_ptrMB(data,ch+1);
    double A=0,B=0,C=0;

    for(i=0;i<size;i++)
      A+=labs(ptrA[i]);
    for(i=0;i<size;i++)
      B+=labs(ptrB[i]);
    for(i=0;i<size;i++)
      C+=labs(ptrA[i]-ptrB[i]);

    if(C<A || C<B)
      if(A<B)
	resultptr[ch>>1]=1;
      else
	resultptr[ch>>1]=2;
    else
      resultptr[ch>>1]=0;
  }
  
  __ogg_appendtoP(dataP,data);
  __ogg_sendtoP(resultP,result);
}

void __oggc_Est(module *m){
  m->action=__ogga_Est;
}

static void __ogga_Esa(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int resultP=__ogg_fixedARG(m,1);

  void *data=__ogg_readfromP(dataP);
  void *result=__ogg_readfromP(resultP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;
  long ressize=__ogg_sizeMB(result);
  long *resptr=__ogg_ptrMB(result,0);

  for(ch=0;ch+1<chan && (ch>>1)<ressize;ch+=2){
    long *ptrA=__ogg_ptrMB(data,ch);
    long *ptrB=__ogg_ptrMB(data,ch+1);
    
    if(resptr[ch>>1])
      if(resptr[ch>>1]==1)
	for(i=0;i<size;i++){
	  ptrB[i]-= ptrA[i];
	  ptrA[i]= -ptrA[i];
	}
      else
	if((resptr[ch>>1]&0x3)==2)
	  for(i=0;i<size;i++){
	    ptrA[i]-= ptrB[i];
	    ptrB[i]= -ptrB[i];
	  }
  }
  
  __ogg_appendtoP(dataP,data);
  __ogg_appendtoP(resultP,result);
}

void __oggc_Esa(module *m){
  m->action=__ogga_Esa;
}

static void __ogga_Epd(module *m){
  int dataP=__ogg_fixedARG(m,0);

  void *data=__ogg_readfromP(dataP);

  int chan=__ogg_chanMB(data),ch;
  long size=__ogg_sizeMB(data),i;

  for(ch=0;ch+1<chan;ch+=2){
    long *ptrA=__ogg_ptrMB(data,ch);
    long *ptrB=__ogg_ptrMB(data,ch+1);
    long A;

    for(i=0;i<size;i++){
      A=ptrA[i];
      ptrA[i]+= ptrB[i];
      ptrB[i]=A-ptrB[i];
    }
  }
  
  __ogg_appendtoP(dataP,data);
}

void __oggc_Epd(module *m){
  m->action=__ogga_Epd;
}

/* ugly hack to be replaced by better native matrix/placement
   support in v99 */

static void __ogga_Epolyq(module *m){
  int inP=__ogg_fixedARG(m,0);
  int outP=__ogg_fixedARG(m,1);
  int windowP=__ogg_fixedARG(m,2);
  float winscale=1./__ogg_fixedARG(m,3);
  int filterP=__ogg_fixedARG(m,4);
  float filtscale=1./__ogg_fixedARG(m,5);
  long step=__ogg_fixedARG(m,6);
  long filters=__ogg_fixedARG(m,7);
  float scale=winscale*filtscale;

  void *in=__ogg_readfromP(inP);
  void *window=__ogg_readfromP(windowP);
  void *filter=__ogg_readfromP(filterP);

  long filtsize=(filters==0?0:__ogg_sizeMB(filter)/filters);
  int chan=__ogg_chanMB(in),ch;

  long size=__ogg_sizeMB(in);
  long winsize=__ogg_sizeMB(window),i;

  long *winptr=__ogg_ptrMB(window,0);
  long *filtptr=__ogg_ptrMB(filter,0);

  /* We don't process the last possible block;
     there are reasons why, but this hack (Epolyq) is
     shortlived so I won't explain.  Work it out. */

  long bandsize=(size-winsize)/step;
  void *out=__ogg_allocMB(chan,bandsize*(filters>0?filters:1));
  float *filtwork=__ogg_malloc(sizeof(float)*filtsize*filters,OM_TEMP);
  float *winwork=__ogg_malloc(sizeof(float)*winsize,OM_TEMP);
  float *work=__ogg_malloc(sizeof(float)*winsize,OM_TEMP);
  int j,k;

  for(i=0;i<winsize;i++)winwork[i]=scale*winptr[winsize-i-1];
  for(i=0;i<filtsize*filters;i++)filtwork[i]=filtptr[i];

  for(ch=0;ch<chan;ch++){
    long *inptr=__ogg_ptrMB(in,ch);
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i*step+winsize<size;i++){
      long *data=inptr+i*step;
      
      /* window.... */
      for(j=0;j<winsize;j++)work[j]=data[j]*winwork[j];
      
      /* ...fold... */
      if(filters==0){

	float accum=0;
	for(k=0;k<winsize;k++)
	  accum+=work[k];
	outptr[i]=rint(accum);

      }else{
	/* We use a symmetry that most MPEG implementations miss that
	   many filterbanks (and all MPEG layers) have.
	   For the v98 demo, we just assume it's there */
	long logicalfiltsize=(filtsize-1)*2;

	for(j=logicalfiltsize;j<winsize;j+=logicalfiltsize)
	  for(k=0;k<logicalfiltsize;k++)
	    work[k]+=work[j+k];

	/* ...and accumulate */
	for(j=0;j<filters;j++){
	  float *onefilt=filtwork+j*filtsize;
	  long offset=j*bandsize;
	  float accum;
	  float *loopptr1=work-1;
	  float *loopptr2=work+logicalfiltsize-1;

	  /* Which brand of symmetry do we have? */
	  if(onefilt[0]==0){
	    /* additive */
	    accum=work[filtsize-2]*onefilt[filtsize-1];

	    for(k=1;k<filtsize-1;k++)
	      accum+=((*(loopptr2-k))+loopptr1[k])*onefilt[k];
	  }else{
	    /* subtractive */
	    accum= (*loopptr2) *onefilt[0];

	    for(k=1;k<filtsize-1;k++)
	      accum+=((*(loopptr2-k))-loopptr1[k])*onefilt[k];
	  }
	  outptr[offset+i]=rint(accum);
	}
      }
    }
  }
  __ogg_sendtoP(inP,in);
  __ogg_sendtoP(outP,out);
  __ogg_sendtoP(filterP,filter);
  __ogg_sendtoP(windowP,window);

  __ogg_free(winwork);
  __ogg_free(filtwork);
  __ogg_free(work);
}

void __oggc_Epolyq(module *m){
  m->action=__ogga_Epolyq;
}


static void __ogga_Esynq(module *m){
  int inP=__ogg_fixedARG(m,0);
  long inshift=__ogg_fixedARG(m,1);
  int outP=__ogg_fixedARG(m,2);
  int windowP=__ogg_fixedARG(m,3);
  long winscale=__ogg_fixedARG(m,4);
  int filterP=__ogg_fixedARG(m,5);
  long filtscale=__ogg_fixedARG(m,6);
  long step=__ogg_fixedARG(m,7);
  long filters=__ogg_fixedARG(m,8);

  void *in=__ogg_readfromP(inP);
  void *window=__ogg_readfromP(windowP);
  void *filter=__ogg_readfromP(filterP);

  long filtsize=(filters==0?0:__ogg_sizeMB(filter)/filters);
  int chan=__ogg_chanMB(in),ch;

  long bandsize=__ogg_sizeMB(in)/filters;
  long winsize=__ogg_sizeMB(window),i;
  long size=winsize+step*(bandsize-1);

  long *winptr=__ogg_ptrMB(window,0);
  long *filtptr=__ogg_ptrMB(filter,0);

  /* We don't process the last possible block;
     there are reasons why, but this hack (Epolyq) is
     shortlived so I won't explain.  Work it out. */

  void *out=__ogg_allocMB(chan,size);

  long *work2=__ogg_malloc(filtsize*sizeof(long),OM_TEMP);
  
  int j,k;

  /* One step at a time.  Mult bands by filters, then window/add */

  for(ch=0;ch<chan;ch++){
    long *inptr=__ogg_ptrMB(in,ch);
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i<bandsize;i++){
      memset(work2,0,sizeof(long)*filtsize);
      for(j=0;j<filters;j++){

	long sample=inptr[j*bandsize+i]>>inshift;
	long *onefilter=filtptr+j*filtsize;

	for(k=0;k<filtsize;k++)
	  work2[k]+=onefilter[k]*sample;

      }

      for(k=0;k<filtsize;k++)
	work2[k]>>=filtscale;
      
      /* window into work[] */
      {
	long *ptr=outptr+i*step;

	for(j=0;j<winsize;){
	  for(k=0;k<filtsize;k++){
	    ptr[j]+=work2[k]*winptr[j];
	    j++;
	  }
	}
      }
    }
    
    for(i=0;i<size;i++)outptr[i]>>=winscale;

  }
  __ogg_sendtoP(inP,in);
  __ogg_sendtoP(outP,out);
  __ogg_sendtoP(filterP,filter);
  __ogg_sendtoP(windowP,window);

  __ogg_free(work2);
}

void __oggc_Esynq(module *m){
  m->action=__ogga_Esynq;
}



