/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: W.c
  function: preset 1D vector functions
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Oct 11 1997

 ********************************************************************/

#include <math.h>
#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "W.h"

static float oPI2 = 6.28318530717958647693;
static float oPI  = 3.14159265358979323846;

static void __ogga_Whl(module *m){
  int ip=__ogg_fixedARG(m,0);
  int amount=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;
  int actual;
  
  if(amount<0)amount=size>>1;
  actual=amount>size?size:amount;
  
  amount++;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<actual;i++)
	ptr[i]*=.5+cos(oPI*(1.-i/(amount-1.)))*.5;
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Whl(module *m){
  m->action=__ogga_Whl;
}

static void __ogga_Whr(module *m){
  int ip=__ogg_fixedARG(m,0);
  int amount=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;
  int actual;
  
  if(amount<0)amount=size>>1;
  actual=amount>size?size:amount;
  
  amount++;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<actual;i++)
	ptr[size-i-1]*=.5+cos(oPI*(1.-i/(amount-1.)))*.5;
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Whr(module *m){
  m->action=__ogga_Whr;
}

static void __ogga_Wsinc(module *m){
  int ip=__ogg_fixedARG(m,0);
  int period=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i,j;
  long chan=__ogg_chanMB(b),ch;

  if(period<=0)period=1;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      float distance=(size&1?1.:.5);
      float dphase=oPI2/period;
      distance*=dphase;
      for(i=size/2-1,j=(size+1)/2;i>=0;i--,j++,distance+=dphase){
	float mul=sin(distance)/distance;
	ptr[i]*=mul;
	ptr[j]*=mul;
      }
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Wsinc(module *m){
  m->action=__ogga_Wsinc;
}

static void __ogga_Wcos(module *m){
  int ip=__ogg_fixedARG(m,0);
  int frequency=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      float mul=(size==0?0:oPI2/size*frequency);
      for(i=0;i<size;i++)
	ptr[i]*=cos(mul*i);
      
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Wcos(module *m){
  m->action=__ogga_Wcos;
}

static void __ogga_Wcos2(module *m){
  int ip=__ogg_fixedARG(m,0);
  int fn=__ogg_fixedARG(m,1);
  int fd=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;

  float frequency=(float)fn/fd;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      float mul=(size==0?0:oPI2/size*frequency);
      for(i=0;i<size;i++)
	ptr[i]*=cos(mul*i);
      
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Wcos2(module *m){
  m->action=__ogga_Wcos2;
}

static void __ogga_Wsin(module *m){
  int ip=__ogg_fixedARG(m,0);
  int frequency=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      float mul=(size==0?0:oPI2/size*frequency);
      for(i=0;i<size;i++)
	ptr[i]*=sin(mul*i);
      
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Wsin(module *m){
  m->action=__ogga_Wsin;
}

static void __ogga_Wsini(module *m){
  int ip=__ogg_fixedARG(m,0);
  int frequency=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      float mul=(size==0?0:oPI2/size*frequency);
      for(i=0;i<size;i++)
	ptr[i]*=-sin(mul*i);
      
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Wsini(module *m){
  m->action=__ogga_Wsini;
}

