#include "stdlib.h"
#include "malloc.h"
#include "daub.h"

/* This code is originally derived from the Daubechies wavelet code in
   _Numerical Recipes in C_ (which has been released to the public
   domain).  The code has been inlined somewhat for faster
   single-threaded performance and changed back to C style 0 indexed
   vectors. */

/* Optimized Daubechies 4th */

#define db4C0  0.4829629131445341
#define db4C1  0.8365163037378079
#define db4C2  0.2241438680420134
#define db4C3 -0.1294095225512604

float *__ogg_daub4f(float *a, float *work, unsigned long n){
  unsigned long nn;
  unsigned long nh,i,j;
  if (n < 4) return (a);
  
  for (nn=n;nn>=4;nn>>=1){
    nh=nn>>1;

    for (i=0,j=0;j<nn-3;j+=2,i++) {
      work[i]=   db4C0*a[j]+db4C1*a[j+1]+db4C2*a[j+2]+db4C3*a[j+3];
      work[i+nh]=db4C3*a[j]-db4C2*a[j+1]+db4C1*a[j+2]-db4C0*a[j+3];
    }
    work[i]=   db4C0*a[nn-2]+db4C1*a[nn-1]+db4C2*a[0]+db4C3*a[1];
    work[i+nh]=db4C3*a[nn-2]-db4C2*a[nn-1]+db4C1*a[0]-db4C0*a[1];

    for (j=0;j<nn;j++) a[j]=work[j];

  }  

  return(a);
}


float *__ogg_daub4i(float *a, float *work, unsigned long n){
  unsigned long nn;
  unsigned long nh,nh1,i,j;
  if (n < 4) return (a);
  
  for (nn=4;nn<=n;nn<<=1){
    nh1=(nh=nn >> 1)+1;

    work[0]=db4C2*a[nh-1]+db4C1*a[nn-1]+db4C0*a[0]+db4C3*a[nh];
    work[1]=db4C3*a[nh-1]-db4C0*a[nn-1]+db4C1*a[0]-db4C2*a[nh];
    for(i=0,j=2;i<nh-1;i++){
      work[j++]=db4C2*a[i]+db4C1*a[i+nh]+db4C0*a[i+1]+db4C3*a[i+nh1];
      work[j++]=db4C3*a[i]-db4C0*a[i+nh]+db4C1*a[i+1]-db4C2*a[i+nh1];
    }

    for (j=0;j<nn;j++) a[j]=work[j];

  }

  return(a);
}

static void pwt(float *a, float *wksp, unsigned long n, int isign, 
		wavefilt *wfilt){
  float ai,ai1;
  unsigned long i,ii,j,f,k,n1,ni,nh;
  float cr[wfilt->ncof];
  
  if (n < 4) return;

  {
    float sig=-1.;
    for(k=0;k<wfilt->ncof;k++){
      cr[wfilt->ncof-k-1]=sig*wfilt->cc[k];
      sig = -sig;
    }
  }

  n1=n-1;
  nh=n >> 1;
  for (j=0;j<n;j++) wksp[j]=0.0;
  if (isign >= 0) {
    for (ii=0,i=0;i<n;i+=2,ii++) {
      ni=i+wfilt->off;
      for (k=0;k<wfilt->ncof;k++) {
	f=n1 & (ni+k);
	wksp[ii] += wfilt->cc[k]*a[f];
	wksp[ii+nh] += cr[k]*a[f];
      }
    }
  } else {
    for (ii=0,i=0;i<n;i+=2,ii++) {
      ai=a[ii];
      ai1=a[ii+nh];
      ni=i+wfilt->off;
      for (k=0;k<wfilt->ncof;k++) {
	f=(n1 & (ni+k));
	wksp[f] += wfilt->cc[k]*ai;
	wksp[f] += cr[k]*ai1;
      }
    }
  }
  for (j=0;j<n;j++) a[j]=wksp[j];
}

void wt1(float *a, float *wksp, unsigned long n, int isign, wavefilt *w){
  unsigned long nn;
  
  if(n<4)return;
  if(isign>=0)
    for (nn=n;nn>=4;nn>>=1) pwt(a,wksp, nn,isign,w);
  else
    for (nn=4;nn<=n;nn<<=1) pwt(a,wksp, nn,isign,w);
}

